#include <pthread.h>
#include "dronedge.h"

static JSClassID jsDroneClassId;

static UA_Boolean pubsubShouldRun = true;
static UA_Boolean pubsubExited = false;

static UA_UInt32 nbDrone;
static JSValueConst *droneObjectIdList;
static MessageQueue messageQueue = {
    .head = NULL,
    .tail = NULL,
};
UA_String currentMessage;

pthread_mutex_t mutex;
pthread_cond_t threadCond;

// Drone class functions

static void js_drone_finalizer(JSRuntime *rt, JSValue val)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(val, jsDroneClassId);
    js_free_rt(rt, s);
}

static JSValue js_drone_ctor(JSContext *ctx, JSValueConst newTarget, 
                             int argc, JSValueConst *argv)
{
    JSDroneData *s;
    JSValue obj = JS_UNDEFINED;
    JSValue proto;
    uint32_t uint32;
    
    s = (JSDroneData *) js_mallocz(ctx, sizeof(*s));
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToUint32(ctx, &uint32, argv[0]))
        goto fail;
    s->id = (uint16_t)uint32;

    proto = JS_GetPropertyStr(ctx, newTarget, "prototype");
    if (JS_IsException(proto))
        goto fail;
    obj = JS_NewObjectProtoClass(ctx, proto, jsDroneClassId);
    JS_FreeValue(ctx, proto);
    if (JS_IsException(obj))
        goto fail;
    JS_SetOpaque(obj, s);
    return obj;
 fail:
    js_free(ctx, s);
    JS_FreeValue(ctx, obj);
    return JS_EXCEPTION;
}

static JSValue js_drone_init(JSContext *ctx, JSValueConst thisVal,
                             int argc, JSValueConst *argv)
{
    int nb;
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &nb, argv[0]))
        return JS_EXCEPTION;
    droneObjectIdList[nb] = thisVal;
    return JS_UNDEFINED;
}

static JSValue js_drone_get(JSContext *ctx, JSValueConst thisVal, int magic)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    JSValue res;
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewUint32(ctx, s->id);
        case 1:
            return JS_NewFloat64(ctx, s->latitude);
        case 2:
            return JS_NewFloat64(ctx, s->longitude);
        case 3:
            return JS_NewFloat64(ctx, s->altitudeAbs);
        case 4:
            return JS_NewFloat64(ctx, s->altitudeRel);
        case 5:
            pthread_mutex_lock(&mutex);
            res = JS_NewString(ctx, s->message);
            strcpy(s->message, "");
            pthread_cond_signal(&threadCond);
            pthread_mutex_unlock(&mutex);
            return res;
        default:
            return JS_EXCEPTION;
    }
}

static JSValue js_drone_set_message(JSContext *ctx, JSValueConst thisVal,
                                    int argc, JSValueConst *argv)
{
    struct messageNode *newNode;
    const char *message;

    message = JS_ToCString(ctx, argv[0]);
    if (strlen(message) > MAX_MESSAGE_SIZE) {
        UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Message too long");
        return JS_EXCEPTION;
    }

    newNode = (struct messageNode*)malloc(sizeof(struct messageNode));
    newNode->message = strdup(message);
    newNode->next = NULL;

    if (messageQueue.tail == NULL) {
        messageQueue.head = messageQueue.tail = newNode;
    } else {
        messageQueue.tail->next = newNode;
        messageQueue.tail = newNode;
    }

    JS_FreeCString(ctx, message);
    return JS_UNDEFINED;
}

static void delete_message_node(struct messageNode *node) {
    free(node->message);
    free(node);
}

static UA_Boolean UA_String_isEmpty(const UA_String *s) {
    return (s->length == 0 || s->data == NULL);
}

static void clear_message(UA_String message) {
    if (!UA_String_isEmpty(&message))
        UA_String_clear(&message); 
}

UA_String get_message(void)
{
    struct messageNode *current;

    current = messageQueue.head;
    if (current == NULL) {
        if (!UA_String_isEmpty(&currentMessage)) {
            UA_String_clear(&currentMessage);
            currentMessage = UA_STRING("");
        }
    } else {
        clear_message(currentMessage);
        currentMessage = UA_STRING_ALLOC(current->message);
        messageQueue.head = current->next == NULL ? (messageQueue.tail = NULL) : current->next;
        delete_message_node(current);
    }
    return currentMessage;
}

static JSClassDef jsDroneClass = {
    "Drone",
    .finalizer = js_drone_finalizer,
};

static const JSCFunctionListEntry js_drone_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("id", js_drone_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("latitude", js_drone_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("longitude", js_drone_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("altitudeAbs", js_drone_get, NULL, 3),
    JS_CGETSET_MAGIC_DEF("altitudeRel", js_drone_get, NULL, 4),
    JS_CGETSET_MAGIC_DEF("message", js_drone_get, NULL, 5),
    JS_CFUNC_DEF("init", 1, js_drone_init),
};

// pubsub functions

UA_UInt16 get_drone_id(UA_UInt32 nb) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    return s->id;
}

VariableData pubsub_get_value(UA_String identifier) {
    VariableData varDetails;
    UA_DataType type;
    UA_Double *array;

    for(UA_UInt32 i = 0; i < countof(droneVariableArray); i++) {
        UA_String name = UA_STRING(droneVariableArray[i].name);
        if(UA_String_equal(&identifier, &name)) {
            varDetails = droneVariableArray[i];
            switch(varDetails.valueRank) {
                case UA_VALUERANK_SCALAR:
                    switch(varDetails.type) {
                        case UA_TYPES_STRING:
                            *(UA_String*)varDetails.value = droneVariableArray[i].getter.getString();
                            break;
                        default:
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "UA_TYPE not handled");
                            break;
                    }
                    break;

                case UA_VALUERANK_ONE_DIMENSION:
                    type = UA_TYPES[varDetails.type];
                    size_t size = varDetails.arrayDimensions[0];
                    array = (UA_Double *) droneVariableArray[i].getter.getArray();

                    if(type.pointerFree) {
                        memcpy(varDetails.value, array, type.memSize * size);
                    } else {
                        uintptr_t ptrs = (uintptr_t)array;
                        uintptr_t ptrd = (uintptr_t)varDetails.value;
                        UA_StatusCode retval = UA_STATUSCODE_GOOD;
                        for(size_t j = 0; j < size; ++j) {
                            retval |= UA_copy((void*)ptrs, (void*)ptrd, &type);
                            ptrs += type.memSize;
                            ptrd += type.memSize;
                        }
                        if(retval != UA_STATUSCODE_GOOD)
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Failed to update array");
                    }

                    free(array);
                    break;

                default:
                    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "UA_VALUERANK not handled");
                    break;
            }
        }
    }

    return varDetails;
}

void init_node_id(UA_UInt32 id, UA_UInt32 nb, UA_UInt32 magic) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    switch(magic) {
        case 0:
            s->positionArrayId = id;
            break;
        case 1:
            s->messageId = id;
            break;
        default:
            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_USERLAND, "Unknown variable id");
            break;
    }
}

void pubsub_update_coordinates(UA_UInt32 id, const UA_DataValue *var)
{
    JSDroneData *s;
    UA_String uaStr;
    UA_Double *positionArray;
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
        s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[i], jsDroneClassId);
        if (s->positionArrayId == id) {
            positionArray = (UA_Double*) var->value.data;
            s->latitude = positionArray[0];
            s->longitude = positionArray[1];
            s->altitudeAbs = positionArray[2];
            s->altitudeRel = positionArray[3];
            return;
        } else if (s->messageId == id) {
            uaStr = *(UA_String*) var->value.data;
            pthread_mutex_lock(&mutex);
            while(strlen(s->message) != 0)
                pthread_cond_wait(&threadCond, &mutex);
            memcpy(s->message, uaStr.data, uaStr.length);
            s->message[uaStr.length] = '\0';
            pthread_mutex_unlock(&mutex);
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

void pubsub_print_coordinates(UA_UInt32 id, const UA_DataValue *var)
{
    JSDroneData *s;
    UA_String uaStr;
    UA_Double *positionArray;
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
        s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[i], jsDroneClassId);
        if (s->positionArrayId == id) {
            positionArray = *(UA_Double**) var->value.data;
            s->latitude = positionArray[0];
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received latitude of drone %d: %f°", s->id, s->latitude);
            s->longitude = positionArray[1];
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received longitude of drone %d: %f°", s->id, s->longitude);
            s->altitudeAbs = positionArray[2];
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received absolute altitude of drone %d: %fm", s->id, s->altitudeAbs);
            s->altitudeRel = positionArray[3];
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received relative altitude of drone %d: %fm", s->id, s->altitudeRel);
            s->altitudeAbs = *(UA_Float*) var->value.data;
            return;
        } else if (s->messageId == id) {
            uaStr = *(UA_String*) var->value.data;
            memcpy(s->message, uaStr.data, uaStr.length);
            s->message[uaStr.length] = '\0';
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received message of drone %d: %s", s->id, s->message);
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

/*
 * arg 0 (string): IPv6 of the multicast group
 * arg 1 (string): port used for multicast communication
 * arg 2 (string): network interface used for multicast communication
 * arg 3 (int): ID of the drone
 * arg 4 (double): publication/subscription interval in ms
 * arg 5 (bool): true if there will be data to publish
*/
static JSValue js_run_pubsub(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    const char *ipv6;
    const char *port;
    const char *netIface;
    char *notConstNetIface;
    char urlBuffer[44];
    UA_UInt32 id;
    UA_Duration interval;
    int res;

    ipv6 = JS_ToCString(ctx, argv[0]);
    port = JS_ToCString(ctx, argv[1]);
    netIface = JS_ToCString(ctx, argv[2]);
    UA_snprintf(urlBuffer, sizeof(urlBuffer), "opc.udp://[%s]:%s/", ipv6, port);

    UA_String transportProfile =
        UA_STRING("http://opcfoundation.org/UA-Profile/Transport/pubsub-udp-uadp");

    UA_NetworkAddressUrlDataType networkAddressUrl;
    notConstNetIface = strdup(netIface);
    networkAddressUrl.networkInterface = UA_STRING(notConstNetIface);
    networkAddressUrl.url              = UA_STRING(urlBuffer);

    if (JS_ToUint32(ctx, &id, argv[3]))
        return JS_EXCEPTION;

    if (JS_ToFloat64(ctx, &interval, argv[4]))
        return JS_EXCEPTION;

    if (JS_ToBool(ctx, argv[5])) {
        res = runPubsub(&transportProfile, &networkAddressUrl,
                        droneVariableArray, countof(droneVariableArray), id,
                        nbDrone, interval, init_node_id, get_drone_id,
                        pubsub_get_value, pubsub_update_coordinates,
                        &pubsubShouldRun);
    } else {
        res = subscribeOnly(&transportProfile, &networkAddressUrl,
                            droneVariableArray, countof(droneVariableArray), id,
                            nbDrone, interval, init_node_id, get_drone_id,
                            pubsub_print_coordinates, &pubsubShouldRun);
    }
    pubsubExited = true;

    JS_FreeCString(ctx, ipv6);
    JS_FreeCString(ctx, port);
    free(notConstNetIface);
    JS_FreeCString(ctx, netIface);
    return JS_NewInt32(ctx, res);
}

static JSValue js_init_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    if (JS_ToUint32(ctx, &nbDrone, argv[0]))
        return JS_EXCEPTION;

    droneObjectIdList = (JSValue *) malloc(nbDrone * sizeof(JSValueConst));
    return JS_NewInt32(ctx, 0);
}

static JSValue js_stop_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    struct messageNode *current;

    pubsubShouldRun = false;
    while(!pubsubExited)
        sleep(1);
    free(droneObjectIdList);

    while (messageQueue.head != NULL) {
        current = messageQueue.head;
        messageQueue.head = current->next;
        delete_message_node(current);
    }
    clear_message(currentMessage);
    return JS_NewInt32(ctx, 0);
}

// Connexion management functions

static JSValue js_mavsdk_start(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    const char *url;
    const char *log_file;
    int timeout;
    int res;

    url = JS_ToCString(ctx, argv[0]);
    log_file = JS_ToCString(ctx, argv[1]);
    if (JS_ToInt32(ctx, &timeout, argv[2]))
        return JS_EXCEPTION;

    res = mavsdk_start(url, log_file, timeout);
    JS_FreeCString(ctx, url);
    JS_FreeCString(ctx, log_file);

    return JS_NewInt32(ctx, res);
}

static JSValue js_mavsdk_stop(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_stop());
}

static JSValue js_mavsdk_reboot(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_reboot());
}

// Flight state management functions

static JSValue js_mavsdk_arm(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_arm());
}

static JSValue js_mavsdk_land(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_land());
}

static JSValue js_mavsdk_takeOff(JSContext *ctx, JSValueConst thisVal,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_takeOff());
}

static JSValue js_mavsdk_takeOffAndWait(JSContext *ctx, JSValueConst thisVal,
                                        int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_takeOffAndWait());
}

static JSValue js_mavsdk_triggerParachute(JSContext *ctx, JSValueConst thisVal,
                                          int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_triggerParachute());
}

// Flight management functions

static JSValue js_mavsdk_loiter(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    double radius;
    if (JS_ToFloat64(ctx, &radius, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_loiter((float)radius));
}

static JSValue js_mavsdk_setAirspeed(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setAirspeed((float)altitude));
}

static JSValue js_mavsdk_setAltitude(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setAltitude((float)altitude));
}

static JSValue js_mavsdk_setTargetCoordinates(JSContext *ctx,
                                              JSValueConst thisVal,
                                              int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;
    double a_arg_double;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &a_arg_double, argv[2]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setTargetCoordinates(la_arg_double,
                                                        lo_arg_double,
                                                        (float)a_arg_double));
}

static JSValue js_mavsdk_setManualControlInput(JSContext *ctx,
                                               JSValueConst thisVal,
                                               int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_setManualControlInput());
}

// Information functions

static JSValue js_mavsdk_getAltitude(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getAltitude());
}

static JSValue js_mavsdk_getAltitudeRel(JSContext *ctx, JSValueConst thisVal,
                                        int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getAltitudeRel());
}

static JSValue js_mavsdk_getInitialAltitude(JSContext *ctx,
                                            JSValueConst thisVal,
                                            int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialAltitude());
}

static JSValue js_mavsdk_getInitialLatitude(JSContext *ctx,
                                            JSValueConst thisVal,
                                            int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialLatitude());
}

static JSValue js_mavsdk_getInitialLongitude(JSContext *ctx,
                                             JSValueConst thisVal,
                                             int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialLongitude());
}

static JSValue js_mavsdk_getLatitude(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getLatitude());
}

static JSValue js_mavsdk_getLongitude(JSContext *ctx, JSValueConst thisVal,
                                      int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getLongitude());
}

static JSValue js_mavsdk_getTakeOffAltitude(JSContext *ctx,
                                            JSValueConst thisVal,
                                            int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getTakeOffAltitude());
}

static JSValue js_mavsdk_getYaw(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getYaw());
}

static JSValue js_mavsdk_getSpeed(JSContext *ctx, JSValueConst thisVal,
                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getSpeed());
}

static JSValue js_mavsdk_healthAllOk(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_healthAllOk());
}

static JSValue js_mavsdk_isInManualMode(JSContext *ctx, JSValueConst thisVal,
                                        int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_isInManualMode());
}

static JSValue js_mavsdk_landed(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_landed());
}

static const JSCFunctionListEntry js_mavsdk_funcs[] = {
    JS_CFUNC_DEF("setMessage", 1, js_drone_set_message ),
    JS_CFUNC_DEF("runPubsub", 6, js_run_pubsub ),
    JS_CFUNC_DEF("stopPubsub", 0, js_stop_pubsub ),
    JS_CFUNC_DEF("start", 3, js_mavsdk_start ),
    JS_CFUNC_DEF("stop", 0, js_mavsdk_stop ),
    JS_CFUNC_DEF("reboot", 0, js_mavsdk_reboot ),
    JS_CFUNC_DEF("arm", 0, js_mavsdk_arm ),
    JS_CFUNC_DEF("land", 0, js_mavsdk_land ),
    JS_CFUNC_DEF("takeOff", 0, js_mavsdk_takeOff ),
    JS_CFUNC_DEF("takeOffAndWait", 0, js_mavsdk_takeOffAndWait ),
    JS_CFUNC_DEF("triggerParachute", 0, js_mavsdk_triggerParachute ),
    JS_CFUNC_DEF("loiter", 1, js_mavsdk_loiter ),
    JS_CFUNC_DEF("setAirspeed", 1, js_mavsdk_setAirspeed ),
    JS_CFUNC_DEF("setAltitude", 1, js_mavsdk_setAltitude ),
    JS_CFUNC_DEF("setTargetCoordinates", 3, js_mavsdk_setTargetCoordinates ),
    JS_CFUNC_DEF("setManualControlInput", 0, js_mavsdk_setManualControlInput ),
    JS_CFUNC_DEF("getAltitude", 0, js_mavsdk_getAltitude ),
    JS_CFUNC_DEF("getAltitudeRel", 0, js_mavsdk_getAltitudeRel ),
    JS_CFUNC_DEF("getInitialAltitude", 0, js_mavsdk_getInitialAltitude ),
    JS_CFUNC_DEF("getInitialLatitude", 0, js_mavsdk_getInitialLatitude ),
    JS_CFUNC_DEF("getInitialLongitude", 0, js_mavsdk_getInitialLongitude ),
    JS_CFUNC_DEF("getLatitude", 0, js_mavsdk_getLatitude ),
    JS_CFUNC_DEF("getLongitude", 0, js_mavsdk_getLongitude ),
    JS_CFUNC_DEF("getTakeOffAltitude", 0, js_mavsdk_getTakeOffAltitude ),
    JS_CFUNC_DEF("getYaw", 0, js_mavsdk_getYaw ),
    JS_CFUNC_DEF("getAirspeed", 0, js_mavsdk_getSpeed ),
    JS_CFUNC_DEF("isInManualMode", 0, js_mavsdk_isInManualMode ),
    JS_CFUNC_DEF("healthAllOk", 0, js_mavsdk_healthAllOk ),
    JS_CFUNC_DEF("landed", 0, js_mavsdk_landed ),
    JS_CFUNC_DEF("initPubsub", 1, js_init_pubsub ),
};

static int js_mavsdk_init(JSContext *ctx, JSModuleDef *m)
{
    JSValue droneProto, droneClass;

    JS_NewClassID(&jsDroneClassId);
    JS_NewClass(JS_GetRuntime(ctx), jsDroneClassId, &jsDroneClass);

    droneProto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, droneProto, js_drone_proto_funcs,
                               countof(js_drone_proto_funcs));

    droneClass = JS_NewCFunction2(ctx, js_drone_ctor, "Drone", 1,
                                   JS_CFUNC_constructor, 0);
    JS_SetConstructor(ctx, droneClass, droneProto);
    JS_SetClassProto(ctx, jsDroneClassId, droneProto);

    JS_SetModuleExport(ctx, m, "Drone", droneClass);

    return JS_SetModuleExportList(ctx, m, js_mavsdk_funcs,
                                  countof(js_mavsdk_funcs));
}

JSModuleDef *js_init_module(JSContext *ctx, const char *module_name)
{
    JSModuleDef *m;
    m = JS_NewCModule(ctx, module_name, js_mavsdk_init);
    if (!m)
        return NULL;
    JS_AddModuleExportList(ctx, m, js_mavsdk_funcs, countof(js_mavsdk_funcs));
    JS_AddModuleExport(ctx, m, "Drone");
    return m;
}
